from unittest import TestCase


# Not tryue anymore
#
# class TestAContractsPerf(TestCase):
#     def test_autoprop_contract_slow(self):
#
#         # TODO : loading PyContracts is extremely slow !!! is there a way to improve ?
#         import cProfile
#         cProfile.run('from contracts.syntax import ParseException')
#         #cProfile.run('from contracts import ContractNotRespected, contract')
#
#         print('this is extremely slow !!!')


from py.xml import html
import pytest
from setuptools_scm.git import GitWorkdir


@pytest.mark.hookwrapper
def pytest_runtest_makereport(item, call):
    """ adds the description field in the report, so that it may be used by the other two functions """
    outcome = yield
    report = outcome.get_result()
    report.description = str(item.function.__doc__)


@pytest.mark.optionalhook
def pytest_html_results_table_header(cells):
    """ Inserts the header for a 'description' column """
    cells[1] = html.th('Test File')
    cells.insert(2, html.th('Test'))
    cells.insert(3, html.th('Description'))
    # cells.insert(0, html.th('Time', class_='sortable time', col='time'))
    # cells.pop()  keep it : we will link to codecov here


@pytest.mark.optionalhook
def pytest_html_results_table_row(report, cells):
    """ Inserts the contents for the 'description' column : the docstring of the test function """
    node_bits = report.nodeid.split('::')
    file = node_bits[0]
    test_id = node_bits[1:]
    line = report.location[1]

    cells[1] = html.th(file)
    cells.insert(2, html.td('::'.join(test_id) if len(node_bits) > 1 else ''))
    cells.insert(3, html.td(report.description))
    # cells.insert(1, html.td(datetime.utcnow(), class_='col-time'))
    # cells.pop()

    # use setuptools_scm to get git hash and build the codecov URL to go and see the source of the test
    # parsed_version = version_from_scm('.')
    # if parsed_version is not None:
    #     hsh = parsed_version.node + ('' if not parsed_version.dirty else '-dirty')

    try:
        hsh = get_git_hash()
        cells[-1] = html.td(html.a('source', href='https://codecov.io/gh/smarie/python-autoclass/src/' + hsh + '/'
                                                  + file + '#L' + str(line)))
    except:
        cells[-1] = html.td('could not read git version')


def get_git_hash():
    # works in local but not on travis
    # proc = subprocess.Popen('git rev-parse --verify --quiet HEAD', stdout=subprocess.PIPE)
    # tmp = proc.stdout.read()
    # rev_node = tmp.decode().replace('\n', '')

    # ep = find_matching_entrypoint('.', 'setuptools_scm.parse_scm')
    # if ep:
    #     ep.load()
    wd = GitWorkdir('.')
    rev_node, _, ret = wd.do_ex('git rev-parse --verify --quiet HEAD')
    return rev_node



import pytest
from autoclass import autoargs


def test_autoargs_simple():
    """ @autoargs Basic functionality, no customization - all constructor arguments are auto-assigned """

    class A(object):
        @autoargs
        def __init__(self, foo, path, debug=False):
            pass

    # Test :
    # -- create an instance
    a = A('rhubarb', 'pie', debug=True)
    # -- check that the fields exist and have the correct value
    assert a.foo == 'rhubarb'
    assert a.path == 'pie'
    assert a.debug == True


def test_autoargs_signature_preserving():
    """ @autoargs Advanced: check that the constructor still has the correct signature """

    class A(object):
        @autoargs
        def __init__(self, foo, path, debug=False):
            pass

    try:
        A()
    except TypeError as e:
        assert e.args[0] == "__init__() missing 2 required positional arguments: 'foo' and 'path'"


def test_autoargs_varargs():
    """
    @autoargs Basic functionality, with special case of variable arguments *args.
    -- note that the variable arguments are stored in a single attribute
    """

    class B(object):
        @autoargs
        def __init__(self, foo, path, debug=False, *args):
            pass

    # Test :
    # -- create an instance
    a = B('rhubarb', 'pie', True, 100, 101)
    # -- check that the fields exist and have the correct value
    assert a.foo == 'rhubarb'
    assert a.path == 'pie'
    assert a.debug == True
    # -- *args is in a single attribute
    assert a.args == (100, 101)


def test_autoargs_varargs_kwvarargs():
    """
    @autoargs Basic functionality, with special case of variable arguments *args and keyword arguments **kw
    -- note that *args are stored in a single attribute while **kw are stored in several attributes
    """

    class C(object):
        @autoargs
        def __init__(self, foo, path, debug=False, *args, **kw):
            pass

    # Test :
    # -- create an instance
    a = C('rhubarb', 'pie', True, 100, 101, verbose=True, bar='bar')
    # -- check that the fields exist and have the correct value
    assert a.foo == 'rhubarb'
    assert a.path == 'pie'
    assert a.debug == True
    # -- *args is in a single attribute
    assert a.args == (100, 101)
    # -- **kw is dispatched in several attributes
    assert a.verbose == True
    assert a.bar == 'bar'


def test_autoargs_noarg():
    """ Same than test_autoargs_simple but with empty arguments list in autoargs """

    class O(object):
        @autoargs()
        def __init__(self, foo, path, debug=False):
            pass

    # Test :
    # -- create an instance
    a = O('rhubarb', 'pie', debug=True)
    # -- check that the fields exist and have the correct value
    assert a.foo == 'rhubarb'
    assert a.path == 'pie'
    assert a.debug == True


def test_autoargs_include():
    """ @autoargs With explicit list of names to include """

    class C(object):
        @autoargs(include=('bar', 'baz', 'verbose'))
        def __init__(self, foo, bar, baz, verbose=False):
            pass

    # Test :
    # -- create an instance
    a = C('rhubarb', 'pie', 1)
    # -- check that the fields exist and have the correct value
    assert a.bar == 'pie'
    assert a.baz == 1
    assert a.verbose == False
    # -- check that a non-included field does not exist
    with pytest.raises(AttributeError):
        print(a.foo)


def test_autoargs_exclude():
    """ @autoargs With explicit list of names to exclude """

    class C(object):
        @autoargs(exclude=('bar', 'baz', 'verbose'))
        def __init__(self, foo, bar, baz, verbose=False):
            pass

    # Test :
    # -- create an instance
    a = C('rhubarb', 'pie', 1)
    # -- check that the fields exist and have the correct value
    assert a.foo == 'rhubarb'
    # -- check that the non-included fields do not exist
    with pytest.raises(AttributeError):
        print(a.bar)
    with pytest.raises(AttributeError):
        print(a.baz)
    with pytest.raises(AttributeError):
        print(a.verbose)


def test_autoargs_include_exclude():
    """ @autoargs assert that you can't use both include/exclude at the same time"""
    with pytest.raises(ValueError):
        class Dummy(object):
            @autoargs(exclude='', include='')
            def __init__(self, foo, bar, baz, verbose=False):
                pass


def test_autoargs_include_exclude_typos():
    """ @autoargs Asserts that errors are correctly raised in case of a nonexistent attribute name in
    include/exclude """

    with pytest.raises(ValueError):
        class Dummy(object):
            @autoargs(exclude='fo')
            def __init__(self, foo, bar, baz, verbose=False):
                pass

    with pytest.raises(ValueError):
        class Dummy(object):
            @autoargs(include='fo')
            def __init__(self, foo, bar, baz, verbose=False):
                pass

# def test_autoargs_exclude_lock():
#     class C(object):
#         @autoargs(exclude=('bar', 'baz', 'verbose'), lock_class_fields=True)
#         def __init__(self, foo, bar, baz, verbose=False):
#             pass
#
#     a = C('rhubarb', 'pie', 1)
#     assert a.foo == 'rhubarb')
#     pytest.raises(AttributeError, getattr, a, 'bar')
#     pytest.raises(AttributeError, setattr, a, 'newa',0)



from typing import Union, Tuple, Type, Any

from autoclass.autoargs import autoargs_decorate
from autoclass.autoprops import autoprops_decorate
from autoclass.autodict import autodict_decorate
from autoclass.utils_reflexion import get_constructor
from autoclass.utils_decoration import _create_class_decorator__robust_to_args


def autoclass(include: Union[str, Tuple[str]]=None, exclude: Union[str, Tuple[str]]=None,
              autoargs: bool=True, autoprops: bool=True, autodict: bool=True):
    """
    A decorator to perform @autoargs, @autoprops and @autodict all at once with the same include/exclude list.

    :param include: a tuple of explicit attributes to include (None means all)
    :param exclude: a named tuple of explicit attributes to exclude. In such case, include should be None.
    :param autoargs: a boolean to enable autoargs on the consturctor (default: True)
    :param autoprops: a boolean to enable autoargs on the consturctor (default: True)
    :param autodict: a boolean to enable autoargs on the consturctor (default: True)
    :return:
    """
    return _create_class_decorator__robust_to_args(autoclass_decorate, include, exclude=exclude, autoargs=autoargs,
                                                   autoprops=autoprops, autodict=autodict)


def autoclass_decorate(cls: Type[Any], include: Union[str, Tuple[str]] = None, exclude: Union[str, Tuple[str]] = None,
                       autoargs: bool=True, autoprops: bool=True, autodict: bool=True) -> Type[Any]:
    """

    :param cls: the class on which to execute. Note that it won't be wrapped.
    :param include: a tuple of explicit attributes to include (None means all)
    :param exclude: a named tuple of explicit attributes to exclude. In such case, include should be None.
    :param autoargs: a boolean to enable autoargs on the consturctor (default: True)
    :param autoprops: a boolean to enable autoargs on the consturctor (default: True)
    :param autodict: a boolean to enable autoargs on the consturctor (default: True)
    :return:
    """

    # @autoargs
    if autoargs:
        init = get_constructor(cls)
        cls.__init__ = autoargs_decorate(init, include=include, exclude=exclude)

    # @autoprops
    if autoprops:
        autoprops_decorate(cls, include=include, exclude=exclude)

    # @autodict
    if autodict:
        autodict_decorate(cls, include=include, exclude=exclude)

    return cls



from typing import List

import pytest
from autoclass import autoargs, autoprops, getter_override, setter_override, \
    IllegalSetterSignatureException, DuplicateOverrideError, autoprops_decorate
from valid8 import Boolean, minlens, gts, between, validate, ValidationError


def test_autoprops_no_contract():
    """ Basic @autoprops functionality, no customization - all constructor arguments become properties """

    @autoprops
    class FooConfigA(object):
        @autoargs
        def __init__(self, a: str, b: List[str]):
            pass

    t = FooConfigA('rhubarb', ['pie', 'pie2'])

    # there are no contracts on the generated setters
    t.a = ''
    t.b = ['r', '']
    # check that the generated getters work
    assert t.a == ''
    assert t.b[0] == 'r'


def test_autoprops_pycontracts():
    """
    @autopropsBasic functionality with PyContracts - if a `@contract` annotation exist on the `__init__` method,
    mentioning a contract for a given parameter, the parameter contract will be added on the generated setter method
    """

    from contracts import ContractNotRespected, contract

    @autoprops
    class FooConfigA(object):
        @autoargs
        @contract(a='str[>0]', b='list[>0](str[>0])')
        def __init__(self, a: str, b: List[str]):
            pass

    t = FooConfigA('rhubarb', ['pie', 'pie2'])

    # check that there are contracts on the generated setters
    with pytest.raises(ContractNotRespected):
        t.a = ''
    with pytest.raises(ContractNotRespected):
        t.b = ['r', '']

    # check that the generated getters work
    t.b = ['r']
    assert t.b[0] == 'r'


def test_autoprops_include():
    """ @autoprops With pycontracts and explicit list of attributes to include """

    from contracts import ContractNotRespected, contract

    @autoprops(include='a')
    class FooConfigB(object):
        @autoargs
        @contract(a='str[>0]', b='list[>0](str[>0])')
        def __init__(self, a: str, b: List[str]):
            pass

    t = FooConfigB('rhubarb', ['pie', 'pie2'])

    # check that there is a contract on the generated setter
    with pytest.raises(ContractNotRespected):
        t.a = ''

    # check that no setter was generated for 'b'
    t.b = ['']  # we can because there is no setter, hence no contract
    assert t.b[0] == ''


def test_autoprops_exclude():
    """ @autoprops With pycontracts and explicit list of attributes to exclude """

    from contracts import ContractNotRespected, contract

    @autoprops(exclude='b')
    class FooConfigB(object):
        @autoargs
        @contract(a='str[>0]', b='list[>0](str[>0])')
        def __init__(self, a: str, b: List[str]):
            pass

    t = FooConfigB('rhubarb', ['pie', 'pie2'])

    # check that there is a contract on the generated setter
    with pytest.raises(ContractNotRespected):
        t.a = ''

    # check that no setter was generated for 'b'
    t.b = ['']  # we can because there is no setter, hence no contract
    assert t.b[0] == ''


def test_autoprops_include_exclude():
    """ @autoprops Asserts that include/exclude cant be used at the same time """

    # you can't use both at the same time
    with pytest.raises(ValueError):
        @autoprops(include='', exclude='')
        class Dummy(object):
            pass


def test_autoprops_override():
    """ @autoprops With Pycontracts. Tests that the user may override generated getter and a setter """

    from contracts import ContractNotRespected, contract

    # check that there is a double-contract warning
    with pytest.warns(UserWarning):
        @autoprops
        class FooConfigC(object):
            @autoargs
            @contract(a='str[>0]', b='list[>0](str[>0])')
            def __init__(self, a: str, b: List[str]):
                pass

            @getter_override
            def a(self):
                # in addition to getting the fields we'd like to print something
                print('a is being read. Its value is \'' + str(self._a) + '\'')
                return self._a

            @setter_override(attribute='b')
            @contract(toto='list[>0](str[>0])')
            def another_name(self, toto: List[str]):
                # in addition to setting the fields we'd like to print something
                print('Property \'b\' was set to \'' + str(toto) + '\'')
                self._b = toto

        t = FooConfigC('rhubarb', ['pie', 'pie2'])

        # check that we can still read a's value
        assert t.a == 'rhubarb'

        # check that 'a' still has a contract on its setter
        with pytest.raises(ContractNotRespected):
            t.a = ''

        # check that 'b' still has a contract on its setter
        with pytest.raises(ContractNotRespected):
            t.b = ['']  # we can not

        # check that 'b' still has a getter generated
        t.b = ['eh', 'oh']
        assert t.b == ['eh', 'oh']


def test_autoprops_override_exceptions():
    """ @autoprops Asserts that the user can not override a generated method if the overriden method has a wrong
    signature """

    # -- a getter is overriden while the attribute does not exist in constructor
    with pytest.raises(AttributeError):
        @autoprops
        class FooConfigD(object):
            @autoargs
            def __init__(self, a: str, b: List[str]):
                pass

            @getter_override
            def c(self):
                return 'c'

    # -- a setter is overriden with a wrong signature
    with pytest.raises(IllegalSetterSignatureException):
        @autoprops
        class FooConfigE(object):
            @autoargs
            def __init__(self, a: str, b: List[str]):
                pass

            @setter_override
            def b(self, toto: List[str], extra_arg: bool = False):
                self._b = toto

    # -- a getter is overriden twice
    with pytest.raises(DuplicateOverrideError):
        @autoprops
        class FooConfigF(object):
            @autoargs
            def __init__(self, a: str, b: List[str]):
                pass

            @getter_override
            def b(self):
                return self._b

            @getter_override(attribute='b')
            def another(self):
                return self._b


def test_autoprops_manual():
    """ @autoprops Tests the manual wrapper autoprops() """

    from contracts import ContractNotRespected
    from contracts import contract

    # we don't use @autoprops here
    class FooConfigA(object):
        @autoargs
        @contract(a='str[>0]', b='list[>0](str[>0])')
        def __init__(self, a: str, b: List[str]):
            pass

    # we execute it here
    autoprops_decorate(FooConfigA)

    t = FooConfigA('rhubarb', ['pie', 'pie2'])

    # check that there are contracts on the generated setters
    with pytest.raises(ContractNotRespected):
        t.a = ''
    with pytest.raises(ContractNotRespected):
        t.b = ['r', '']

    # check that the generated getters work
    t.b = ['r']
    assert t.b[0] == 'r'


def test_autoprops_signature_preserving():
    """ @autoprops Advanced: checks that the generated constructor still has the correct signature"""

    @autoprops
    class FooConfigD(object):
        @autoargs
        def __init__(self, a: str, b: List[str]):
            pass

    t = FooConfigD('rhubarb', ['pie', 'pie2'])
    try:
        getattr(FooConfigD, 'a').fset(t)
    except TypeError as e:
        assert e.args[0] == "generated_setter_fun() missing 1 required positional argument: 'val'"

    try:
        getattr(FooConfigD, 'a').fget()
    except TypeError as e:
        # yes ; right now it is still a lambda, since using a named method does not seem to work :(
        assert e.args[0] == "<lambda>() missing 1 required positional argument: 'self'"


def test_autoprops_enforce_validate():
    """ Makes sure that autoprops works with enforce AND validate """

    # from autoclass import autoargs, autoprops, Boolean
    import enforce as en
    from enforce import runtime_validation
    from numbers import Real, Integral
    from typing import Optional

    en.config(dict(mode='covariant'))  # allow subclasses when validating types

    @runtime_validation
    @autoprops
    class HouseConfiguration(object):
        @autoargs
        @validate(name=minlens(0),
                  surface=gts(0),
                  nb_floors=between(1, 100, open_right=True))
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            pass

        # -- overriden setter for surface for custom validation or other things
        @setter_override
        def surface(self, surface):
            print('Set surface to {}'.format(surface))
            self.toto = 'done'
            self._surface = surface

    t = HouseConfiguration('test', 12, 2)

    # 'Optional' works
    t.nb_floors = None

    # Custom print works
    t.surface = 0.1
    assert t.toto == 'done'

    # Type validation works
    from enforce.exceptions import RuntimeTypeError
    with pytest.raises(RuntimeTypeError):
        t.nb_floors = 2.2

    # Custom validation works
    with pytest.raises(ValidationError):
        t.surface = 0


def test_autoprops_enforce_default():
    """ Tests that the default value is also set in the setters if it is provided in the constructor """

    from enforce import runtime_validation

    @runtime_validation
    @autoprops
    class Foo:
        def __init__(self, mandatory_field: str, optional_field: str = None):
            pass

    a = Foo('t')
    a.optional_field = None  # raises RuntimeTypeError: Argument 'val' was not of type <class 'str'>. Actual type was NoneType.



from inspect import getfullargspec
from typing import Tuple, Callable, Union

from autoclass.utils_decoration import _create_function_decorator__robust_to_args
from autoclass.utils_include_exclude import _sieve
from autoclass.var_checker import check_var
from decorator import decorate


def autoargs(include: Union[str, Tuple[str]]=None, exclude: Union[str, Tuple[str]]=None):
    """
    Defines a decorator with parameters, to automatically assign the inputs of a function to self PRIOR to executing 
    the function. In other words:
    
    ```
    @autoargs
    def myfunc(a):
        print('hello')
    ```
    
    will create the equivalent of
    
    ```
    def myfunc(a):
        self.a = a
        print('hello')
    ```
    
    Initial code from http://stackoverflow.com/questions/3652851/what-is-the-best-way-to-do-automatic-attribute-assignment-in-python-and-is-it-a#answer-3653049

    :param include: a tuple of attribute names to include in the auto-assignment. If None, all arguments will be
    included by default
    :param exclude: a tuple of attribute names to exclude from the auto-assignment. In such case, include should be None
    :return:
    """
    return _create_function_decorator__robust_to_args(autoargs_decorate, include, exclude=exclude)


def autoargs_decorate(func: Callable, include: Union[str, Tuple[str]]=None, exclude: Union[str, Tuple[str]]=None) \
        -> Callable:
    """
    Defines a decorator with parameters, to automatically assign the inputs of a function to self PRIOR to executing 
    the function. This is the inline way to apply the decorator 
    
    ```
    myfunc2 = autoargs_decorate(myfunc)
    ```
    
    
    See autoargs for details.

    :param func: the function to wrap
    :param include: a tuple of attribute names to include in the auto-assignment. If None, all arguments will be
    included by default
    :param exclude: a tuple of attribute names to exclude from the auto-assignment. In such case, include should be None
    :return:
    """

    # (0) first check parameters
    if include is not None and exclude is not None:
        raise ValueError('Only one of \'include\' or \'exclude\' argument should be provided.')
    check_var(include, var_name='include', var_types=[str, tuple], enforce_not_none=False)
    check_var(exclude, var_name='exclude', var_types=[str, tuple], enforce_not_none=False)

    # (1) then retrieve function signature
    # attrs, varargs, varkw, defaults = getargspec(func)
    signature_attrs, signature_varargs, signature_varkw, signature_defaults, signature_kwonlyargs, \
    signature_kwonlydefaults, signature_annotations = getfullargspec(func)
    # TODO better use signature(func) ? see how it's done in valid8

    # check that include/exclude dont contain names that are incorrect
    if include is not None:
        incorrect = set(list([include] if isinstance(include, str) else include)) - set(signature_attrs)
        if len(incorrect) > 0:
            raise ValueError('@autoargs definition exception: include contains \'' + str(incorrect) + '\' that is/are '
                            'not part of signature for ' + str(func))
    if exclude is not None:
        incorrect = set([exclude] if isinstance(exclude, str) else exclude) - set(signature_attrs)
        if len(incorrect) > 0:
            raise ValueError('@autoargs definition exception: exclude contains \'' + str(incorrect) + '\' that is/are '
                             'not part of signature for ' + str(func))

    # (2) Optionally lock the class only for the provided fields
    # Does not work for the moment. Besides locking fields seems to have issues with pickle serialization
    # so we'd rather not propose this option.
    #
    # See 'attrs' project for this kind of advanced features https://github.com/python-attrs/attrs
    #
    # if lock_class_fields:
    #     if signature_varkw:
    #         raise Exception('cant lock field names with variable kwargs')
    #     else:
    #         object_type = get_class_that_defined_method(func)
    #         if include:
    #             fields = include
    #         else:
    #             fields = signature_attrs[1:]
    #             if signature_varargs:
    #                 fields.append(signature_varargs)
    #             if exclude:
    #                 for a in exclude:
    #                     fields.remove(a)
    #
    #         # right now, doesnot work
    #         _lock_fieldnames_class(object_type, field_names=tuple(fields))

    # (3) Finally, create a wrapper around the function so that all attributes included/not excluded are
    # set to self BEFORE executing the function.

    # -- old:
    # @functools.wraps(func) -> to make the wrapper function look like the wrapped function
    # def wrapper(self, *args, **kwargs):

    # -- new:
    # we now use 'decorate' at the end of this code to have a wrapper that has the same signature, see below
    def wrapper(func, self, *args, **kwargs):

        # handle default values and kw arguments
        for attr, val in zip(reversed(signature_attrs), reversed(signature_defaults or [])):
            if sieve(attr):
                # set default or provided value
                if attr in kwargs.keys():
                    # provided: we never seem to enter here, why ? maybe depends on the version of python
                    setattr(self, attr, kwargs[attr])
                else:
                    # default
                    setattr(self, attr, val)

        # handle positional arguments except 'self' (the first)
        signature_positional_attrs = signature_attrs[1:len(args)+1]
        for attr,val in zip(signature_positional_attrs,args):
            if sieve(attr):
                setattr(self, attr, val)

        # handle varargs : since we dont know their name, store them in a global field named after the vararg name
        if signature_varargs:
            remaining_args=args[len(signature_positional_attrs):]
            if sieve(signature_varargs):
                setattr(self, signature_varargs, remaining_args)

        # handle varkw : since we know their names, store them directly in independent fields
        if signature_varkw:
            for attr, val in kwargs.items():
                if sieve(signature_varkw) or sieve(attr):
                    setattr(self, attr, val)

        # finally execute the function
        return func(self, *args, **kwargs)

    def sieve(attr):
        """
        Locally-defined function that we use in the wrapper to check if an attribute shall be included in the processing
        :param attr: 
        :return: 
        """
        return _sieve(attr, include=include, exclude=exclude)

    # return wrapper
    return decorate(func, wrapper)



from inspect import signature
from typing import Type, Any, Tuple, Union, Dict
from warnings import warn

from autoclass import check_var
from autoclass.utils_include_exclude import _sieve
from autoclass.utils_reflexion import get_constructor
from autoclass.utils_decoration import _create_class_decorator__robust_to_args, _check_known_decorators

from collections import Mapping


def autodict(include: Union[str, Tuple[str]]=None, exclude: Union[str, Tuple[str]]=None,
             only_constructor_args: bool = True, only_public_fields: bool = False):
    """
    A decorator to makes objects of the class behave like a read-only `dict`. It does two things:
    * it adds collections.Mapping to the list of parent classes (i.e. to the class' `__bases__`)
    * it generates `__len__`, `__iter__` and `__getitem__` in order for the appropriate fields to be exposed in the dict
    view.
    * it adds a static from_dict method to build objects from dicts (only if only_constructor_args=True)
    * it overrides eq method if not already implemented
    * it overrides str and repr method if not already implemented
    
    Parameters allow to customize the list of fields that will be visible.

    :param include: a named tuple of explicit attributes to include (None means all)
    :param exclude: a named tuple of explicit attributes to exclude. In such case, include should be None.
    :param only_constructor_args: if True (default), only constructor arguments will be exposed through the dictionary
    view, not any other field that would be created in the constructor or dynamically. This makes it very convenient
    to use in combination with @autoargs. If set to False, the dictionary is a direct view of public object fields.
    :param only_public_fields: this parameter is only used when only_constructor_args is set to False. If
    only_public_fields is set to False, all fields are visible. Otherwise (default), class-private fields will be hidden
    :return:
    """
    return _create_class_decorator__robust_to_args(autodict_decorate, include, exclude=exclude,
                                                   only_constructor_args=only_constructor_args,
                                                   only_public_fields=only_public_fields)


def autodict_decorate(cls: Type[Any], include: Union[str, Tuple[str]] = None,
                      exclude: Union[str, Tuple[str]] = None, only_constructor_args: bool = True,
                      only_public_fields: bool = False) -> Type[Any]:
    """
    To automatically generate the appropriate methods so that objects of this class behave like a `dict`,
    manually, without using @autodict decorator.

    :param cls: the class on which to execute. Note that it won't be wrapped.
    :param include: a named tuple of explicit attributes to include (None means all)
    :param exclude: a named tuple of explicit attributes to exclude. In such case, include should be None.
    :param only_constructor_args: if True (default), only constructor arguments will be exposed through the dictionary
    view, not any other field that would be created in the constructor or dynamically. This makes it very convenient
    to use in combination with @autoargs. If set to False, the dictionary is a direct view of public object fields.
    :param only_public_fields: this parameter is only used when only_constructor_args is set to False. If
    only_public_fields is set to False, all fields are visible. Otherwise (default), class-private fields will be hidden
    :return:
    """

    # first check that we do not conflict with other known decorators
    _check_known_decorators(cls, '@autodict')

    # perform the class mod
    _execute_autodict_on_class(cls, include=include, exclude=exclude, only_constructor_args=only_constructor_args,
                               only_public_fields=only_public_fields)

    return cls


def _execute_autodict_on_class(object_type: Type[Any], include: Union[str, Tuple[str]]=None,
                               exclude: Union[str, Tuple[str]]=None, only_constructor_args: bool = True,
                               only_public_fields: bool = True):
    """
    This method makes objects of the class behave like a read-only `dict`. It does two things:
    * it adds collections.Mapping to the list of parent classes (i.e. to the class' `__bases__`)
    * it generates `__len__`, `__iter__` and `__getitem__` in order for the appropriate fields to be exposed in the dict
    view.
    * it adds a static from_dict method to build objects from dicts (only if only_constructor_args=True)
    * it overrides eq method if not already implemented
    * it overrides str and repr method if not already implemented

    Parameters allow to customize the list of fields that will be visible.

    :param object_type: the class on which to execute.
    :param include: a named tuple of explicit attributes to include (None means all)
    :param exclude: a named tuple of explicit attributes to exclude. In such case, include should be None.
    :param only_constructor_args: if True (default), only constructor arguments will be exposed through the dictionary
    view, not any other field that would be created in the constructor or dynamically. This makes it very convenient
    to use in combination with @autoargs. If set to False, the dictionary is a direct view of public object fields.
    :param only_public_fields: this parameter is only used when only_constructor_args is set to False. If
    only_public_fields is set to False, all fields are visible. Otherwise (default), class-private fields will be hidden
    :return:
    """

    if include is not None and exclude is not None:
        raise ValueError('Only one of \'include\' or \'exclude\' argument should be provided.')
    check_var(include, var_name='include', var_types=[str, tuple], enforce_not_none=False)
    check_var(exclude, var_name='exclude', var_types=[str, tuple], enforce_not_none=False)

    if issubclass(object_type, Mapping):
        raise ValueError('@autodict can not be set on classes that are already subclasses of Mapping, and therefore '
                         'already behave like dict')

    # 1. implement the abstract method required by Mapping to work
    if hasattr(object_type, '__len__'):
        warn('__len__ is already defined on this class, it will be overriden with the one generated by '
             '@autodict/@autoclass !')
    if hasattr(object_type, '__iter__'):
        warn('__iter__ is already defined on this class, it will be overriden with the one generated by '
             '@autodict/@autoclass !')
    if hasattr(object_type, '__getitem__'):
        warn('__getitem__ is already defined on this class, it will be overriden with the one generated by '
             '@autodict/@autoclass !')

    # Construct the Mapping methods according to the options
    if only_constructor_args:
        # ** easy: we know the exact list of fields to make visible in the Mapping
        # a. Find the __init__ constructor signature
        constructor = get_constructor(object_type)
        s = signature(constructor)

        # b. Collect all attributes that are not 'self' and are included and not excluded
        added = []
        for attr_name in s.parameters.keys():
            if _sieve(attr_name, include=include, exclude=exclude):
                added.append(attr_name)

        # c. Finally build the methods
        def __iter__(self):
            """
            Generated by @autodict.
            Implements the __iter__ method from collections.Iterable by relying on a hardcoded list of fields
            :param self:
            :return:
            """
            return iter(added)

        def __len__(self):
            """
            Generated by @autodict.
            Implements the __len__ method from collections.Sized by relying on a hardcoded list of fields
            :param self:
            :return:
            """
            return len(added)

        def __getitem__(self, key):
            """
            Generated by @autodict.
            Implements the __getitem__ method from collections.Mapping by relying on a hardcoded list of fields
            :param self:
            :param key:
            :return:
            """
            if key in added:
                try:
                    return getattr(self, key)
                except AttributeError:
                    raise KeyError('@autodict generated dict view - invalid field name: ' + key)
            else:
                raise KeyError('@autodict generated dict view - invalid or hidden field name: ' + key)

    else:
        # ** all dynamic fields are allowed
        if include is None and exclude is None and not only_public_fields:
            # easy: all of vars is exposed
            def __iter__(self):
                """
                Generated by @autodict.
                Implements the __iter__ method from collections.Iterable by relying on vars(self)
                :param self:
                :return:
                """
                return iter(att_name for att_name in vars(self))

            def __len__(self):
                """
                Generated by @autodict.
                Implements the __len__ method from collections.Sized by relying on vars(self)
                :param self:
                :return:
                """
                return len(vars(self))

            def __getitem__(self, key):
                """
                Generated by @autodict.
                Implements the __getitem__ method from collections.Mapping by relying on getattr(self, key)
                :param self:
                :param key:
                :return:
                """
                try:
                    return getattr(self, key)
                except AttributeError:
                    raise KeyError('@autodict generated dict view - invalid field name: ' + key)
        else:
            # harder: dynamic filters
            # private_name_prefix = '_' + object_type.__name__ + '_'
            private_name_prefix = '_'

            def __iter__(self):
                """
                Generated by @autodict.
                Implements the __iter__ method from collections.Iterable by relying on a filtered vars(self)
                :param self:
                :return:
                """
                for att_name in vars(self):
                    if _sieve(att_name, include=include, exclude=exclude):
                        if not only_public_fields \
                                or (only_public_fields and not att_name.startswith(private_name_prefix)):
                            yield att_name

            def __len__(self):
                """
                Generated by @autodict.
                Implements the __len__ method from collections.Sized by relying on a filtered vars(self)
                :param self:
                :return:
                """
                # rely on iter()
                return sum(1 for e in self)

            def __getitem__(self, key):
                """
                Generated by @autodict.
                Implements the __getitem__ method from collections.Mapping by relying on a filtered getattr(self, key)
                :param self:
                :param key:
                :return:
                """
                if hasattr(self, key):
                    if _sieve(key, include=include, exclude=exclude):
                        if not only_public_fields \
                                or (only_public_fields and not key.startswith(private_name_prefix)):
                            return getattr(self, key)
                        else:
                            raise KeyError('@autodict generated dict view - hidden field name: ' + key)
                else:
                    raise KeyError('@autodict generated dict view - invalid field name: ' + key)

    object_type.__len__ = __len__
    object_type.__iter__ = __iter__
    object_type.__getitem__ = __getitem__

    # 2. add the methods from Mapping to the class
    # -- current proposition: add inheritance dynamically
    bazz = tuple(t for t in object_type.__bases__ if t is not object)
    if len(bazz) == len(object_type.__bases__):
        object_type.__bases__ = bazz + (Mapping,)
    else:
        object_type.__bases__ = bazz + (Mapping, object)

        # -- alternate way: add methods one by one
        # meths = getmembers(Mapping, predicate=callable)
        # for name, func in meths:
        #     if name != '__getitem__':
        #         # bind method to this class too (we access 'im_func' to get the original method)
        #         setattr(object_type, name, func.im_func)

    # 3. add the static class method to build objects from dict, if needed
    if only_constructor_args:

        def from_dict(cls, dct: Dict):
            """
            Generated by @autodict.
            A class method to construct an object from a dictionary of field values.

            :param cls:
            :param dct:
            :return:
            """
            return cls(**dct)

        object_type.from_dict = classmethod(from_dict)

    # 4. override equality method if not already implemented
    if '__eq__' not in object_type.__dict__:

        def __eq__(self, other):
            """
            Generated by @autodict.
            In the case the other is of the same type, use the dict comparison. Otherwise, falls back to super.

            :param self:
            :param other:
            :return:
            """
            # in the case the other is of the same type, use the dict comparison
            if isinstance(other, object_type):
                return dict(self) == other
            else:
                # else fallback to inherited behaviour, whatever it is
                return super(object_type, self).__eq__(other)

        object_type.__eq__ = __eq__

    # 5. override str and repr method if not already implemented
    if '__str__' not in object_type.__dict__:

        def __str__(self):
            """
            Generated by @autodict.
            Uses the dict representation and puts the type in front

            :param self:
            :return:
            """
            return type(self).__name__ + '(' + str(dict(self)) + ')'

        object_type.__str__ = __str__

    if '__repr__' not in object_type.__dict__:
        def __repr__(self):
            """
            Generated by @autodict.
            Uses the dict representation and puts the type in front
            maybe?

            :param self:
            :return:
            """
            return type(self).__name__ + '(' + str(dict(self)) + ')'

        object_type.__repr__ = __repr__

    return



import pytest
from autoclass import autoclass, setter_override, autoprops, autoargs

from autoclass.utils_decoration import AutoclassDecorationException
from valid8 import Boolean


def test_autoclass_enforce_validate_not_reversed():
    """ Tests that if we reverse the annotations orders, it still works. Currently it fails """

    from autoclass import autoclass, setter_override
    from numbers import Real
    from enforce import runtime_validation, config
    config(dict(mode='covariant'))  # to accept subclasses in validation

    @runtime_validation
    @autoclass
    class HouseConfiguration(object):
        def __init__(self, surface: Real):
            pass

        # -- overriden setter for surface
        @setter_override
        def surface(self, surface):
            print('Set surface to {}'.format(surface))
            self._surface = surface

    t = HouseConfiguration(12)


def test_autoclass_enforce_validate_reversed():
    """ Tests that if we reverse the annotations orders, it still works. Currently it fails """

    from autoclass import autoclass, setter_override
    from numbers import Real
    from enforce import runtime_validation, config
    config(dict(mode='covariant'))  # to accept subclasses in validation

    with pytest.raises(AutoclassDecorationException):
        @autoclass
        @runtime_validation
        class HouseConfiguration(object):
            def __init__(self, surface: Real):
                pass

            # -- overriden setter for surface
            @setter_override
            def surface(self, surface):
                print('Set surface to {}'.format(surface))
                self._surface = surface

        t = HouseConfiguration(12)


def test_readme_pytypes():
    """ Makes sure that the code in the documentation page is correct for the pytypes example """

    # from autoclass import autoargs, autoprops, Boolean
    from pytypes import typechecked
    from numbers import Real, Integral
    from typing import Optional

    @typechecked
    @autoclass
    class HouseConfiguration(object):
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            pass

        # -- overriden setter for surface for custom validation
        @setter_override
        def surface(self, surface):
            assert surface > 0
            self._surface = surface

    t = HouseConfiguration('test', 12, 2)

    # 'Optional' works
    t.nb_floors = None

    # Type validation works
    from pytypes import InputTypeError
    with pytest.raises(InputTypeError):
        t.nb_floors = 2.2

    # Custom validation works
    with pytest.raises(AssertionError):
        t.surface = 0


def test_readme_enforce():
    """ Makes sure that the code in the documentation page is correct for the enforce example """

    # from autoclass import autoargs, autoprops, Boolean
    import enforce as en
    from enforce import runtime_validation
    from numbers import Real, Integral
    from typing import Optional

    en.config(dict(mode='covariant'))  # allow subclasses when validating types

    @runtime_validation
    @autoprops
    class HouseConfiguration(object):
        @autoargs
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            pass

        # -- overriden setter for surface for custom validation
        @setter_override
        def surface(self, surface):
            assert surface > 0
            self._surface = surface

    t = HouseConfiguration('test', 12, 2)

    # 'Optional' works
    t.nb_floors = None

    # Type validation works
    from enforce.exceptions import RuntimeTypeError
    with pytest.raises(RuntimeTypeError):
        t.nb_floors = 2.2

    # Custom validation works
    with pytest.raises(AssertionError):
        t.surface = 0



import pytest
from typing import List

from autoclass import autodict


@pytest.mark.parametrize('only_public_fields', [True, False], ids=lambda x: 'only_public' if x else 'including class-private dunder fields')
@pytest.mark.parametrize('only_constructor_args', [True, False], ids=lambda x: 'only_constructor_args' if x else 'all_obj_fields')
def test_autodict(only_constructor_args, only_public_fields):
    """ Basic @autodict functionality, no customization - all constructor fields become visible in the resulting
    dictionary """

    @autodict(only_constructor_args=only_constructor_args, only_public_fields=only_public_fields)
    class FooConfigA(object):

        dummy_class_field = 'just to be sure it does not appear'

        def __init__(self, a: str, b: List[str]):
            self.a = a
            self.b = b
            self.c = 't'
            self._weak_private = 'r'
            self.__class_private = 't'

        def dummy_func(self):
            """ we create this just to be sure the function is not in the dict view """
            pass

    t = FooConfigA('rhubarb', ['pie', 'pie2'])
    t.new_field = 0
    t._new_field_weak_private = 1
    t.__new_field_class_private_incorrect = 0

    class Dummy:
        t.__new_field_class_private = 1

    # check that the dict view works
    # --constructor arguments
    assert t['a'] == t.a
    assert t['b'] == t.b

    # --other 'static' fields
    if only_constructor_args:
        with pytest.raises(KeyError):
            assert t['c'] == t.c
    else:
        assert t['c'] == t.c

    if only_constructor_args or only_public_fields:
        with pytest.raises(KeyError):
            assert t['_weak_private'] == t._weak_private
        with pytest.raises(KeyError):
            assert t['_FooConfigA__class_private'] == t._FooConfigA__class_private
    else:
        assert t['_weak_private'] == t._weak_private
        assert t['_FooConfigA__class_private'] == t._FooConfigA__class_private

    # -- dynamic fields
    if only_constructor_args:
        # new fields should not appear
        with pytest.raises(KeyError):
            assert t['new_field'] == t.new_field
    else:
        # new fields should appear
        assert t['new_field'] == t.new_field

    if only_constructor_args or only_public_fields:
        # new private fields should not appear
        with pytest.raises(KeyError):
            assert t['_new_field_weak_private'] == t._new_field_weak_private
        with pytest.raises(KeyError):
            assert t['__new_field_class_private_incorrect'] == t.__new_field_class_private_incorrect
        with pytest.raises(KeyError):
            assert t['_Dummy__new_field_class_private'] == t._Dummy__new_field_class_private
    else:
        # new private fields should appear
        assert t['_new_field_weak_private'] == t._new_field_weak_private
        assert t['__new_field_class_private_incorrect'] == t.__new_field_class_private_incorrect
        assert t['_Dummy__new_field_class_private'] == t._Dummy__new_field_class_private

    # assert that the dict is read-only
    with pytest.raises(TypeError):
        t['x'] = 5

    # assert that equals works
    assert t == dict(t)

    if only_constructor_args:
        # we could use t but we use dict(t) so that the error message in pytest is more readable
        assert dict(t) == {'a': 'rhubarb',  # only the two constructor fields appear
                           'b': ['pie', 'pie2']}
    elif only_public_fields:
        # we could use t but we use dict(t) so that the error message in pytest is more readable
        assert dict(t) == {'a': 'rhubarb',
                           'b': ['pie', 'pie2'],
                           'c': 't',
                           # _FooConfigA__class_private should not appear
                           'new_field': 0,
                           #'_weak_private': 'r',
                           #'_new_field_weak_private': 1,
                           # private fields defined out of the objects class are still visible
                           #'__new_field_class_private_incorrect': 0,
                           #'_Dummy__new_field_class_private': 1
                           }
    else:
        # we could use t but we use dict(t) so that the error message in pytest is more readable
        assert dict(t) == {'a': 'rhubarb',
                           'b': ['pie', 'pie2'],
                           'c': 't',
                           '_FooConfigA__class_private': 't',  # <= this is the one private field that appears now
                           'new_field': 0,
                           '_weak_private': 'r',
                           '_new_field_weak_private': 1,
                           # private fields defined out of the objects class are still visible
                           '__new_field_class_private_incorrect': 0,
                           '_Dummy__new_field_class_private': 1}

    # assert that the generated static method works
    if only_constructor_args:
        assert FooConfigA.from_dict(t) == t



from typing import Union, Tuple


def _sieve(attr, include:Union[str, Tuple[str]]=None, exclude:Union[str, Tuple[str]]=None):
    """decide whether an action has to be performed on the attribute or not, based on its name"""

    if include is not None and exclude is not None:
        raise ValueError('Only one of \'include\' or \'exclude\' argument should be provided.')

    # win time by not doing this
    # check_var(include, var_name='include', var_types=[str, tuple], enforce_not_none=False)
    # check_var(exclude, var_name='exclude', var_types=[str, tuple], enforce_not_none=False)

    if attr is 'self':
        return False
    if exclude and attr in exclude:
        return False
    if not include or attr in include:
        return True
    else:
        return False


from typing import Callable, Union, List, Set

"""
This var checker is only provided here as an example of what you'd like to do when you dont use any contract library
"""

class MissingMandatoryParameterException(Exception):
    """ This is raised whenever a mandatory parameter is missing or null/None"""


def check_var(var, var_types:Union[type, List[type]] =None, var_name=None, enforce_not_none:bool = True,
              allowed_values:Set = None, min_value = None, min_strict:bool = False,
              max_value = None, max_strict:bool = False, min_len:int = None, min_len_strict:bool = False,
              max_len:int = None, max_len_strict:bool = False):
    """
    Helper method to check that an object has certain properties:
    * not none
    * a certain type
    * in some accepted values
    * in some accepted range

    :param var: the object to check
    :param var_types: the type(s) to enforce. If None, type will not be enforced
    :param var_name: the name of the varioable to be used in error messages
    :param enforce_not_none: boolean, default True. Whether to enforce that var is not None.
    :param allowed_values: an optional set of allowed values
    :param min_value: an optional minimum value
    :param min_strict: if True, only values strictly greater than the minimum value will be accepted
    :param max_value: an optional maximum value
    :param max_strict: if True, only values strictly lesser than the minimum value will be accepted
    :return:
    """

    var_name = var_name or 'object'

    if enforce_not_none and (var is None):
        # enforce not none
        raise MissingMandatoryParameterException('Error, ' + var_name + '" is mandatory, it should be non-None')

    if not (var is None) and not (var_types is None):
        # enforce type
        if not isinstance(var_types, list):
            var_types = [var_types]

        match = False
        for var_type in var_types:
            # just in case, even though users should use FunctionType or MethodType which is the true type
            if var_type is Callable:
                if callable(var):
                    match = True
                    break
            else:
                if isinstance(var, var_type):
                    match = True
                    break

        if not match:
            raise TypeError('Error, ' + var_name + '" should be one of type(s) ' + str(var_types) + ', found: ' + str(type(var)))

    if var is not None:
        if allowed_values is not None:
            # enforce allowed values
            if var not in allowed_values:
                raise TypeError('Error, ' + var_name + '" should be one of "' + str(allowed_values) + '", found: ' + str(var))

        if min_value is not None:
            # enforce min value
            if min_strict:
                if not (var > min_value):
                    raise TypeError(
                        'Error, ' + var_name + '" should be strictly greater than "' + str(min_value) + '", found: ' + str(var))
            else:
                if not (var >= min_value):
                    raise TypeError(
                        'Error, ' + var_name + '" should be greater than "' + str(min_value) + '", found: ' + str(var))

        if max_value is not None:
            # enforce max value
            if max_strict:
                if not (var < max_value):
                    raise TypeError(
                        'Error, ' + var_name + '" should be strictly lesser than "' + str(max_value) + '", found: ' + str(var))
            else:
                if not (var <= max_value):
                    raise TypeError(
                        'Error, ' + var_name + '" should be lesser than "' + str(max_value) + '", found: ' + str(var))

        if min_len is not None:
            # enforce min length
            if min_len_strict:
                if not (len(var) > min_len):
                    raise TypeError(
                        'Error, ' + var_name + '" length should be strictly greater than "' + str(min_len) + '", found: ' + str(len(var)))
            else:
                if not (len(var) >= min_len):
                    raise TypeError(
                        'Error, ' + var_name + '" length should be greater than "' + str(min_len) + '", found: ' + str(len(var)))

        if max_len is not None:
            # enforce max length
            if max_len_strict:
                if not (len(var) < max_len):
                    raise TypeError(
                        'Error, ' + var_name + '" length should be strictly lesser than "' + str(max_len) + '", found: ' + str(len(var)))
            else:
                if not (len(var) <= max_len):
                    raise TypeError(
                        'Error, ' + var_name + '" length should be lesser than "' + str(max_len) + '", found: ' + str(len(var)))



from inspect import signature, getmembers, Parameter
from typing import Type, Any


# def get_missing_mandatory_parameters(param_names:set, object_type:Type[Any]):
#     """
#     Utility method to get the set of missing mandatory parameters, from a set of parameters and an object type
#
#     :param param_names:
#     :param object_type:
#     :return:
#     """
#     return set(param_names) - set(get_mandatory_param_names(object_type))


# def get_all_param_names(item_type: Type[Any]):
#     """
#     Utility function to extract the constructor and find all its parameter names
#
#     :param item_type:
#     :return:
#     """
#     # extract unique constructor signature
#     constructor = get_constructor(item_type)
#     s = signature(constructor)
#
#     # return all parameters
#     return [attribute_name for attribute_name, param in s.parameters.items()]


# def get_mandatory_param_names(item_type: Type[Any]):
#     """
#     Utility function to extract the constructor and find its mandatory parameter names
#
#     :param item_type:
#     :return:
#     """
#     # extract unique constructor signature
#     constructor = get_constructor(item_type)
#     s = signature(constructor)
#
#     # return mandatory parameters
#     return [attribute_name for attribute_name, param in s.parameters.items() if param.default is Parameter.empty]


def get_constructor(typ, allow_inheritance: bool=False):
    """
    Utility method to return the unique constructor (__init__) of a type

    :param typ: a type
    :param allow_inheritance: if True, the constructor will be returned even if it is not defined in this class
    (inherited). By default this is set to False: an exception is raised when no constructor is explicitly defined in
    the class
    :return: the found constructor
    """

    # constructors = [f[1] for f in getmembers(typ) if f[0] is '__init__']
    # if len(constructors) > 1:
    #     raise Exception('Several constructors were found for class ' + str(typ))
    # if len(constructors) == 0:
    #     raise Exception('No constructor was found for class ' + str(typ))
    #
    # constructor = constructors[0]
    #
    # # if constructor is a wrapped function, access to the underlying function

    # faster: just access it!
    if allow_inheritance:
        return typ.__init__
    else:
        # check that the constructor is really defined here
        if '__init__' in typ.__dict__:
            return typ.__init__
        else:
            raise Exception('No explicit constructor was found for class ' + str(typ))


# def get_class_that_defined_method(meth):
#     """
#     Utility method, from http://stackoverflow.com/questions/3589311/get-defining-class-of-unbound-method-object-in-python-3/25959545#25959545
#
#     :param meth:
#     :return:
#     """
#
#     if ismethod(meth):
#         # this won't happen in python 3.5 : https://bugs.python.org/issue27901
#         for cls in getmro(meth.__self__.__class__):
#            if cls.__dict__.get(meth.__name__) is meth:
#                 return cls
#         meth = meth.__func__ # fallback to __qualname__ parsing
#     if isfunction(meth):
#         cls = getattr(getmodule(meth),
#                       meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0])
#         if isinstance(cls, type):
#             return cls
#     return None # not required since None would have been implicitly returned anyway


# def _lock_fieldnames_class(object_type: Type[Any], field_names:Tuple[str]=None):
#     """
#     Utility method to lock the possible fields of a class to the arguments declared in the constructor
#
#     :param object_type:
#     :param field_names: optional explicit list of field names
#     :return:
#     """
#
#     if not field_names:
#
#         # 1. Find the constructor
#         # extract unique constructor __init__
#         constructor = get_constructor(object_type)
#
#         # extract the __init__ signature
#         s = signature(constructor)
#
#         field_names = s.parameters.keys()
#
#     # now lock the names of fields : no additional field can be created on this object anymore
#     object_type.__slots__ = field_names



# allow users to do
#     from autoclass import xxx
from autoclass.autoargs import *
from autoclass.autoclass import *
from autoclass.autoprops import *
from autoclass.autodict import *
from autoclass.var_checker import *

# allow users to do
#     import autoclass as ac
__all__ = ['autoclass', 'autoargs', 'autoprops', 'autodict', 'var_checker']



import pytest
from autoclass import autoargs, autoprops, setter_override, autodict, autoclass
from typing import List, Tuple, Dict, Any
from valid8 import Boolean, minlens, validate, ValidationError, gt, is_in, minlen


def test_readme_enforce_validate():
    """ Makes sure that the code in the documentation page is correct for the enforce + validate example """

    from autoclass import autoclass, setter_override
    from valid8 import Boolean, validate, minlens, gt
    from numbers import Real, Integral
    from typing import Optional

    # we use enforce runtime checker for this example
    from enforce import runtime_validation, config
    config(dict(mode='covariant'))  # to accept subclasses in validation

    @runtime_validation
    @autoclass
    class HouseConfiguration(object):

        @validate(name=minlens(0),
                  surface=gt(0))
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            pass

        # -- overriden setter for surface for custom validation or other things
        @setter_override
        def surface(self, surface):
            print('Set surface to {}'.format(surface))
            self._surface = surface

    t = HouseConfiguration('test', 12, 2)

    # 'Optional' works
    t.nb_floors = None

    # Type validation works
    from enforce.exceptions import RuntimeTypeError
    with pytest.raises(RuntimeTypeError):
        t.nb_floors = 2.2

    # Value validation works
    with pytest.raises(ValidationError):
        t.surface = -1

    # Value validation works in constructor
    with pytest.raises(ValidationError):
        HouseConfiguration('', 12, 2)

    # str and repr work
    assert str(t) == repr(t)
    # "HouseConfiguration({'nb_floors': None, 'with_windows': False, 'surface': 12, 'name': 'test'})"
    # assert eval(repr(t)) == t does not work !

    # dict work
    assert t == {'name': 'test', 'nb_floors': None, 'surface': 12, 'with_windows': False}
    assert t == dict(name='test', nb_floors=None, surface=12, with_windows=False)
    t.keys()
    for k, v in t.items():
        print(str(k) + ': ' + str(v))
    HouseConfiguration.from_dict({'name': 'test2', 'surface': 1})


def test_readme_old_way():
    """ Makes sure that the code in the documentation page is correct for the 'old way' of writing classes """

    from autoclass import check_var
    from numbers import Real, Integral
    from typing import Optional, Union
    from valid8 import Boolean

    class HouseConfiguration(object):
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            self.name = name
            self.surface = surface
            self.nb_floors = nb_floors
            self.with_windows = with_windows

        # --name
        @property
        def name(self):
            return self._name

        @name.setter
        def name(self, name: str):
            check_var(name, var_name='name', var_types=str)
            self._name = name

        # --surface
        @property
        def surface(self) -> Real:
            return self._surface

        @surface.setter
        def surface(self, surface: Real):
            check_var(surface, var_name='surface', var_types=Real, min_value=0, min_strict=True)
            self._surface = surface

        # --nb_floors
        @property
        def nb_floors(self) -> Optional[Integral]:
            return self._nb_floors

        @nb_floors.setter
        def nb_floors(self, nb_floors: Optional[Integral]):
            check_var(nb_floors, var_name='nb_floors', var_types=Integral, enforce_not_none=False)
            self._surface = nb_floors  # !**

        # --with_windows
        @property
        def with_windows(self) -> Boolean:
            return self._with_windows

        @with_windows.setter
        def with_windows(self, with_windows: Boolean):
            check_var(with_windows, var_name='with_windows', var_types=Boolean)
            self._with_windows = with_windows

    HouseConfiguration('test', 0.1)


def test_readme_pycontracts():
    """ Makes sure that the code in the documentation page is correct for the PyContracts example """

    from contracts import contract, ContractNotRespected
    from numbers import Real, Integral
    from typing import Optional

    @autoprops
    class HouseConfiguration(object):
        @autoargs
        @contract(name='str[>0]',
                  surface='(int|float),>=0',
                  nb_floors='None|int',
                  with_windows='bool')
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            pass

        # -- overriden setter for surface - no need to repeat the @contract
        @setter_override
        def surface(self, surface: Real):
            assert surface > 0
            self._surface = surface

    t = HouseConfiguration('test', 0.1)
    t.nb_floors = None
    with pytest.raises(ContractNotRespected):
        t.nb_floors = 2.2
    with pytest.raises(ContractNotRespected):
        t.surface = -1


@pytest.mark.skip(reason="open bug in pytypes https://github.com/Stewori/pytypes/issues/19")
def test_readme_pytypes_validate():
    """ Makes sure that the code in the documentation page is correct for the pytypes + validate example """

    from autoclass import autoclass, setter_override
    from valid8 import Boolean, validate, minlens, gt
    from numbers import Real, Integral
    from typing import Optional

    # we use pytypes for this example
    from pytypes import typechecked

    @typechecked
    @autoclass
    class HouseConfiguration(object):

        @validate(name=minlens(0),
                  surface=gt(0))
        def __init__(self,
                     name: str,
                     surface: Real,
                     nb_floors: Optional[Integral] = 1,
                     with_windows: Boolean = False):
            pass

        # -- overriden setter for surface for custom validation or other things
        @setter_override
        def surface(self, surface):
            print('Set surface to {}'.format(surface))
            self._surface = surface

    t = HouseConfiguration('test', 12, 2)

    # 'Optional' works
    t.nb_floors = None

    # Type validation works
    from pytypes import InputTypeError
    with pytest.raises(InputTypeError):
        t.nb_floors = 2.2

    # Value validation works
    with pytest.raises(ValidationError):
        t.surface = -1

    # Value validation works in constructor
    with pytest.raises(ValidationError):
        HouseConfiguration('', 12, 2)

    # str and repr work
    assert str(t) == repr(t)
    # "HouseConfiguration({'nb_floors': None, 'with_windows': False, 'surface': 12, 'name': 'test'})"
    # assert eval(repr(t)) == t does not work !

    # dict work
    assert t == {'name': 'test', 'nb_floors': None, 'surface': 12, 'with_windows': False}
    assert t == dict(name='test', nb_floors=None, surface=12, with_windows=False)
    t.keys()
    for k, v in t.items():
        print(str(k) + ': ' + str(v))

    # TODO this is an open bug in pytypes https://github.com/Stewori/pytypes/issues/19
    HouseConfiguration.from_dict({'name': 'test2', 'surface': 1})


def test_readme_usage_autoprops_validate():
    @autoprops
    class FooConfigA(object):
        @autoargs
        @validate(a=minlens(0))
        def __init__(self, a: str):
            pass

    t = FooConfigA('rhubarb')

    # check that the generated getters work
    t.a = 'r'
    assert t.a == 'r'

    # check that there are validators on the generated setters
    with pytest.raises(ValidationError):
        t.a = ''  # raises ValidationError


def test_readme_usage_autodict_1():
    """ basic autodict without and with autoargs """

    # ** without autoargs
    @autodict
    class A(object):
        def __init__(self, a: int, b: str):
            self.a = a
            self.b = b

    o = A(1, 'r')
    # o behaves like a read-only dict
    assert o == dict(o)
    assert o == {'a': 1, 'b': 'r'}

    # you can create an object from a dict too thanks to the generated class function
    p = A.from_dict({'a': 1, 'b': 'r'})
    assert p == o

    # str and repr methods show interesting stuff
    str(p)  # "A({'a': 1, 'b': 'r'})"
    repr(p)  # "A({'a': 1, 'b': 'r'})"

    # ** with autoargs
    @autodict
    class B(object):
        @autoargs
        def __init__(self, a: int, b: str):
            pass

    o = B(1, 'r')
    # same results
    assert o == {'a': 1, 'b': 'r'}

    # you can create an object from a dict too thanks to the generated class function
    p = B.from_dict({'a': 1, 'b': 'r'})
    assert p == o


def test_readme_usage_autodict_2():
    """ basic autodict with other and private fields """

    @autodict
    class C(object):
        @autoargs
        def __init__(self, a: str, b: List[str]):
            self.non_constructor_arg = 't'
            self._private = 1
            self.__class_private = 't'

    o = C(1, 'r')
    # only fields corresponding to constructor arguments are visible
    assert o == {'a': 1, 'b': 'r'}


def test_readme_usage_autodict_3():

    @autodict(only_constructor_args=False, only_public_fields=False)
    class D(object):
        @autoargs
        def __init__(self, a: str, b: List[str]):
            self.non_constructor_arg = 'b'
            self._private = 1
            self.__class_private = 't'

    o = D(1, 'r')
    # o behaves like a read-only dict, all fields are now visible
    assert o == dict(o)
    assert o == {'a': 1, 'b': 'r',
                 'non_constructor_arg': 'b',
                 '_private': 1,
                 '_D__class_private': 't'}  # notice the name


def test_readme_usage_autoclass():

    from numbers import Integral
    from typing import Optional

    # we will use enforce as the runtime checker
    import enforce as en
    from enforce import runtime_validation
    en.config(dict(mode='covariant'))  # allow subclasses when validating types

    # class definition
    @runtime_validation
    @autoclass
    class AllOfTheAbove:
        @validate(a=gt(1), c=minlen(1))
        def __init__(self, a: Integral, b: Boolean, c: Optional[List[str]] = None):
            pass

    # instance creation
    o = AllOfTheAbove(a=2, b=True)

    # @autoargs works
    assert o.a == 2

    # @autoprops works, in combination with any runtime checker (here demonstrated with enforce)
    from enforce.exceptions import RuntimeTypeError
    with pytest.raises(RuntimeTypeError):
        o.b = 1  # RuntimeTypeError Argument 'b' was not of type Boolean. Actual type was int.

    # @autodict works
    assert o == {'a': 2, 'b': True, 'c': None}
    assert AllOfTheAbove.from_dict(o) == o
    assert dict(**o) == o


def test_readme_usage_autoclass_custom():

    from numbers import Integral
    from typing import Optional

    # we will use enforce as the runtime checker
    import enforce as en
    from enforce import runtime_validation
    en.config(dict(mode='covariant'))  # allow subclasses when validating types

    # class definition
    @runtime_validation
    @autoclass(autodict=False)
    class PartsOfTheAbove:
        @validate(a=gt(1), c=minlen(1))
        def __init__(self, a: Integral, b: Boolean, c: Optional[List[str]] = None):
            pass

    # instance creation
    o = PartsOfTheAbove(a=2, b=True)

    # @autoargs works
    assert o.a == 2

    # @autoprops works, in combination with any runtime checker (here demonstrated with enforce)
    from enforce.exceptions import RuntimeTypeError
    with pytest.raises(RuntimeTypeError):
        o.b = 1  # RuntimeTypeError Argument 'b' was not of type Boolean. Actual type was int.

    # @autodict is disabled
    with pytest.raises(AssertionError):
        assert o == {'a': 2, 'b': True, 'c': None}  # AssertionError
    with pytest.raises(AttributeError):
        assert PartsOfTheAbove.from_dict(o) == o  # AttributeError: type object 'PartsOfTheAbove' has no attribute 'from_dict'
    with pytest.raises(TypeError):
        assert dict(**o) == o  # TypeError: type object argument after ** must be a mapping, not PartsOfTheAbove



from inspect import getmembers, signature, Parameter
from typing import Type, Any, Tuple, Callable, Union, Optional
from warnings import warn

from decorator import decorate

from autoclass import check_var
from autoclass.utils_include_exclude import _sieve
from autoclass.utils_reflexion import get_constructor
from autoclass.utils_decoration import _create_function_decorator__robust_to_args, \
    _create_class_decorator__robust_to_args, _check_known_decorators

__GETTER_OVERRIDE_ANNOTATION = '__getter_override__'
__SETTER_OVERRIDE_ANNOTATION = '__setter_override__'


class IllegalGetterSignatureException(Exception):
    """ This is raised whenever an overridden getter has an illegal signature"""


class IllegalSetterSignatureException(Exception):
    """ This is raised whenever an overridden setter has an illegal signature"""


class DuplicateOverrideError(Exception):
    """ This is raised whenever a getter or setter is overridden twice for the same attribute"""


def autoprops(include: Union[str, Tuple[str]]=None, exclude: Union[str, Tuple[str]]=None):
    """
    A decorator to automatically generate all properties getters and setters from the class constructor.
    * if a @contract annotation exist on the __init__ method, mentioning a contract for a given parameter, the
    parameter contract will be added on the generated setter method
    * The user may override the generated getter and/or setter by creating them explicitly in the class and annotating
    them with @getter_override or @setter_override. Note that the contract will still be dynamically added on the
    setter, even if the setter already has one (in such case a `UserWarning` will be issued)

    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :return:
    """
    return _create_class_decorator__robust_to_args(autoprops_decorate, include, exclude=exclude)


def autoprops_decorate(cls: Type[Any], include: Union[str, Tuple[str]] = None,
                       exclude: Union[str, Tuple[str]] = None) -> Type[Any]:
    """
    To automatically generate all properties getters and setters from the class constructor manually, without using
    @autoprops decorator.
    * if a @contract annotation exist on the __init__ method, mentioning a contract for a given parameter, the
    parameter contract will be added on the generated setter method
    * The user may override the generated getter and/or setter by creating them explicitly in the class and annotating
    them with @getter_override or @setter_override. Note that the contract will still be dynamically added on the
    setter, even if the setter already has one (in such case a `UserWarning` will be issued)

    :param cls: the class on which to execute. Note that it won't be wrapped.
    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :return:
    """

    # first check that we do not conflict with other known decorators
    _check_known_decorators(cls, '@autoprops')

    # perform the class mod
    _execute_autoprops_on_class(cls, include=include, exclude=exclude)

    # TODO better create a wrapper than modify the class? Probably not
    # class Autoprops_Wrapper(object):
    #     def __init__(self, *args, **kwargs):
    #         self.wrapped = cls(*args, **kwargs)
    #
    # return Autoprops_Wrapper

    return cls


def _execute_autoprops_on_class(object_type: Type[Any], include: Union[str, Tuple[str]]=None,
                                exclude: Union[str, Tuple[str]]=None):
    """
    This method will automatically add one getter and one setter for each constructor argument, except for those
    overridden using autoprops_override_decorate(), @getter_override or @setter_override.
    It will add a @contract on top of all setters (generated or overridden, if they don't already have one)

    :param object_type: the class on which to execute.
    :param include: a tuple of explicit attribute names to include (None means all)
    :param exclude: a tuple of explicit attribute names to exclude. In such case, include should be None.
    :return:
    """

    if include is not None and exclude is not None:
        raise ValueError('Only one of \'include\' or \'exclude\' argument should be provided.')
    check_var(include, var_name='include', var_types=[str, tuple], enforce_not_none=False)
    check_var(exclude, var_name='exclude', var_types=[str, tuple], enforce_not_none=False)

    # 1. Find the __init__ constructor signature and possible pycontracts @contract
    constructor = get_constructor(object_type)
    s = signature(constructor)

    # option a) pycontracts
    contracts_dict = constructor.__contracts__ if hasattr(constructor, '__contracts__') else {}
    # option b) validate
    validators_dict = constructor.__validators__ if hasattr(constructor, '__validators__') else {}

    # 2. For each attribute that is not 'self' and is included and not excluded, add the property
    added = []
    for attr_name in s.parameters.keys():
        if _sieve(attr_name, include=include, exclude=exclude):
            added.append(attr_name)
            _add_property(object_type, s.parameters[attr_name],
                          pycontract=(contracts_dict[attr_name] if attr_name in contracts_dict.keys() else None),
                          validators=(validators_dict[attr_name] if attr_name in validators_dict.keys() else None))

    # 3. Finally check that there is no overridden setter or getter that does not correspond to an attribute
    extra_overrides = getmembers(object_type, predicate=(lambda fun: callable(fun) and
                                                                     (hasattr(fun, __GETTER_OVERRIDE_ANNOTATION)
                                                                      and getattr(fun, __GETTER_OVERRIDE_ANNOTATION) not in added)
                                                                      or
                                                                     (hasattr(fun, __SETTER_OVERRIDE_ANNOTATION))
                                                                      and getattr(fun, __SETTER_OVERRIDE_ANNOTATION) not in added)
                                 )
    if len(extra_overrides) > 0:
        raise AttributeError('Attribute named \'' + extra_overrides[0][0] + '\' was not found in constructor signature.'
                             'Therefore its getter/setter can not be overridden by function '
                             + extra_overrides[0][1].__qualname__)


def _add_property(object_type: Type[Any], parameter: Parameter, pycontract: Any = None, validators: Any = None):
    """
    A method to dynamically add a property to a class with the optional given pycontract or validators.
    If the property getter and/or setter have been overridden, it is taken into account too.

    :param object_type: the class on which to execute.
    :param parameter:
    :param pycontract:
    :param validators:
    :return:
    """
    property_name = parameter.name

    # 1. create the private field name , e.g. '_foobar'
    private_property_name = '_' + property_name

    # 2. property getter (@property) - create or use overridden
    getter_fun = _get_getter_fun(object_type, parameter, private_property_name)

    # 3. property setter (@property_name.setter) - create or use overridden
    setter_fun, var_name = _get_setter_fun(object_type, parameter, private_property_name)

    # 4. add the contract to the setter, if any
    setter_fun_with_possible_contract = setter_fun
    if pycontract is not None:
        setter_fun_with_possible_contract = _add_contract_to_setter(setter_fun, var_name, pycontract,
                                                                    property_name)
    elif validators is not None:
        setter_fun_with_possible_contract = _add_validators_to_setter(setter_fun, var_name, validators,
                                                                      property_name)

    # 5. change the function name to make it look nice
    setter_fun_with_possible_contract.__name__ = property_name
    setter_fun_with_possible_contract.__module__ = object_type.__module__
    setter_fun_with_possible_contract.__qualname__ = object_type.__name__ + '.' + property_name
    # __annotations__
    # __doc__
    # __dict__

    # 6. Finally add the property to the class
    # WARNING : property_obj.setter(f) does absolutely nothing :) > we have to assign the result
    # setattr(object_type, property_name, property_obj.setter(f))
    new_prop = property(fget=getter_fun, fset=setter_fun_with_possible_contract)

    # # specific for enforce: here we might wrap the overriden property setter on which enforce has already written
    # # something.
    # if hasattr(setter_fun_with_possible_contract, '__enforcer__'):
    #     new_prop.__enforcer__ = setter_fun_with_possible_contract.__enforcer__
    # DESIGN DECISION > although this would probably work, it is probably better to 'force' users to always use the
    # @autoprops annotation BEFORE any other annotation. This is now done in autoprops_decorate

    setattr(object_type, property_name, new_prop)

    return


def _has_annotation(annotation, value):
    """ Returns a function that can be used as a predicate in get_members, that  """

    def matches_property_name(fun):
        """ return true if fun is a callable that has the correct annotation with value """
        return callable(fun) and hasattr(fun, annotation) \
               and getattr(fun, annotation) is value
    return matches_property_name


def _get_getter_fun(object_type: Type, parameter: Parameter, private_property_name: str):
    """
    Utility method to find the overridden getter function for a given property, or generate a new one

    :param object_type:
    :param property_name:
    :param private_property_name:
    :return:
    """

    property_name = parameter.name

    # -- check overridden getter for this property name
    overridden_getters = getmembers(object_type, predicate=_has_annotation(__GETTER_OVERRIDE_ANNOTATION, property_name))

    if len(overridden_getters) > 0:
        if len(overridden_getters) > 1:
            raise DuplicateOverrideError('Getter is overridden more than once for attribute name : ' + property_name)

        # --use the overridden getter
        getter_fun = overridden_getters[0][1]

        # --check its signature
        s = signature(getter_fun)
        if not ('self' in s.parameters.keys() and len(s.parameters.keys()) == 1):
            raise IllegalGetterSignatureException('overridden getter must only have a self parameter, found ' +
                                                  str(len(s.parameters.items()) - 1) + ' for function ' + str(
                getter_fun.__qualname__))

        # --use the overridden getter
        property_obj = property(getter_fun)
    else:
        # -- generate the getter :
        # @property
        # def foobar(self):
        #     return self.__foobar

        # --for some reason, this does not work, but the lambda function works !
        # def generated_getter_fun(self):
        #     getattr(self, private_property_name)
        #
        # # -- use the generated getter
        # getter_fun = generated_getter_fun

        getter_fun = lambda self: getattr(self, private_property_name)
        getter_fun.__annotations__['return'] = parameter.annotation  # add type hint to output declaration

    return getter_fun


def _get_setter_fun(object_type: Type, parameter: Parameter, private_property_name: str):
    """
    Utility method to find the overridden setter function for a given property, or generate a new one

    :param object_type:
    :param property_name:
    :param property_type:
    :param private_property_name:
    :return:
    """

    property_name = parameter.name

    overridden_setters = getmembers(object_type, _has_annotation(__SETTER_OVERRIDE_ANNOTATION, property_name))

    if len(overridden_setters) > 0:
        # --check that we only have one
        if len(overridden_setters) > 1:
            raise DuplicateOverrideError('Setter is overridden more than once for attribute name : ' + property_name)

        # --use the overridden setter
        setter_fun = overridden_setters[0][1]

        # --find the parameter name and check the signature
        s = signature(setter_fun)
        p = [attribute_name for attribute_name, param in s.parameters.items() if attribute_name is not 'self']
        if len(p) != 1:
            raise IllegalSetterSignatureException('overridden setter must have only 1 non-self argument, found ' +
                                                  str(len(s.parameters.items()) - 1) + ' for function '
                                                  + str(setter_fun.__qualname__))
        var_name = p[0]

    else:
        # --create the setter, equivalent of:
        # @foobar.setter
        # def foobar(self, foobar):
        #     self.__foobar = foobar

        # the setter function
        def generated_setter_fun(self, val):
            return setattr(self, private_property_name, val)

        # remember the parameter name - we dont need to change it in the signature
        var_name = 'val'

        # add type hint annotation if provided
        if parameter.annotation:
            generated_setter_fun.__annotations__['val'] = parameter.annotation

        # add default value if provided
        if parameter.default is not Parameter.empty:
            generated_setter_fun.__defaults__ = (parameter.default,)

        # other options (more complex) to fiddle with the signature :
        # http://stackoverflow.com/questions/18625510/how-can-i-programmatically-change-the-argspec-of-a-function-not-in-a-python-de
        # http://stackoverflow.com/questions/10303248/true-dynamic-and-anonymous-functions-possible-in-python

        setter_fun = generated_setter_fun

    return setter_fun, var_name


def _add_contract_to_setter(setter_fun, var_name, property_contract, property_name):

    # 0. check that we can import contracts
    try:
        # noinspection PyUnresolvedReferences
        from contracts import ContractNotRespected, contract
    except ImportError as e:
        raise Exception('Use of _add_contract_to_setter requires that PyContract library is installed. Check that you '
                        'can \'import contracts\'')

    # -- check if a contract already exists on the function
    if hasattr(setter_fun, '__contracts__'):
        msg = 'overridden setter for attribute ' + property_name + ' implemented by function ' \
              + str(setter_fun.__qualname__) + ' has a contract while there is a contract already defined ' \
              + 'for this property in the __init__ constructor. This will lead to double-contract in the final ' \
              + 'setter, please remove the one on the overridden setter.'
        warn(msg)

    # -- add the generated contract
    setter_fun_with_possible_contract = contract(setter_fun, **{var_name: property_contract})

    # the only thing we can't do is to replace the function's parameter name dynamically in the error messages
    # so we wrap the function again to catch the potential pycontracts error :(
    # old:
    # @functools.wraps(func) -> to make the wrapper function look like the wrapped function
    # def wrapper(self, *args, **kwargs):
    # new:
    # we now use 'decorate' to have a wrapper that has the same signature, see below
    def _contracts_parser_interceptor(func, self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except ContractNotRespected as er:
            er.error = er.error.replace('\'val\'', '\'' + property_name + '\'')
            raise er

    # f = _contracts_parser_interceptor(f)
    setter_fun_with_possible_contract = decorate(setter_fun_with_possible_contract, _contracts_parser_interceptor)
    return setter_fun_with_possible_contract


def _add_validators_to_setter(setter_fun, var_name, validators, property_name):

    # 0. check that we can import validate
    # note: this is useless now but maybe one day validate will be another project ?
    try:
        # noinspection PyUnresolvedReferences
        from valid8 import decorate_with_validators
    except ImportError:
        raise Exception('Use of _add_contract_to_setter requires that validate library is installed. Check that you can'
                        ' \'import validate\'')

    # -- check if a contract already exists on the function
    if hasattr(setter_fun, '__validators__'):
        msg = 'overridden setter for attribute ' + property_name + ' implemented by function ' \
              + str(setter_fun.__qualname__) + ' has validators while there are validators already defined ' \
              'for this property in the __init__ constructor. This will lead to double-contract in the final ' \
              'setter, please remove the one on the overridden setter.'
        warn(msg)

    # -- add the generated contract
    setter_fun_with_validation = decorate_with_validators(setter_fun, **{var_name: validators})

    # # the only thing we can't do is to replace the function's parameter name dynamically in the validation error
    # #  messages so we wrap the function again to catch the potential pycontracts error :(
    # # old:
    # # @functools.wraps(func) -> to make the wrapper function look like the wrapped function
    # # def wrapper(self, *args, **kwargs):
    # # new:
    # # we now use 'decorate' to have a wrapper that has the same signature, see below
    # def _contracts_parser_interceptor(func, self, *args, **kwargs):
    #     try:
    #         return func(self, *args, **kwargs)
    #     except ContractNotRespected as e:
    #         e.error = e.error.replace('\'val\'', '\'' + property_name + '\'')
    #         raise e

    # f = _contracts_parser_interceptor(f)
    # setter_fun_with_possible_contract = decorate(setter_fun_with_possible_contract, _contracts_parser_interceptor)
    return setter_fun_with_validation


def getter_override(attribute:str = None):
    """
    A decorator to indicate an overridden getter for a given attribute. If the attribute name is None, the function name
    will be used as the attribute name.

    :param attribute: the attribute name for which the decorated function is an overridden getter
    :return:
    """
    return _create_function_decorator__robust_to_args(autoprops_override_decorate, attribute, is_getter=True)


def setter_override(attribute:str = None):
    """
    A decorator to indicate an overridden setter for a given attribute. If the attribute name is None, the function name
     will be used as the attribute name. The @contract will still be dynamically added.

    :param attribute: the attribute name for which the decorated function is an overridden setter
    :return:
    """
    return _create_function_decorator__robust_to_args(autoprops_override_decorate, attribute, is_getter=False)


def autoprops_override_decorate(func: Callable, attribute:str = None, is_getter:bool = True) -> Callable:
    """
    Used to decorate a function as an overridden getter or setter, without using the @getter_override or
    @setter_override annotations. If the overridden setter has no @contract, the contract will still be
     dynamically added. Note: this should be executed BEFORE @autoprops or autoprops_decorate().

    :param func: the function on which to execute. Note that it won't be wrapped but simply annotated.
    :param attribute: the attribute name. If None, the function name will be used
    :param is_getter: True for a getter override, False for a setter override.
    :return:
    """

    # Simply annotate the fact that this is a function
    attr_name = attribute or func.__name__
    if is_getter:
        if hasattr(func, __GETTER_OVERRIDE_ANNOTATION):
            raise DuplicateOverrideError('Getter is overridden twice for attribute name : ' + attr_name)
        else:
            func.__getter_override__ = attr_name
    else:
        if hasattr(func, __SETTER_OVERRIDE_ANNOTATION):
            raise DuplicateOverrideError('Setter is overridden twice for attribute name : ' + attr_name)
        else:
            func.__setter_override__ = attr_name

    return func



from inspect import isclass
from typing import Callable


class AutoclassDecorationException(Exception):
    pass


def _check_known_decorators(typ, calling_decorator: str) -> bool:
    """
    Checks that a given type is not already decorated by known decorators that may cause trouble.
    If so, it raises an Exception
    :return:
    """
    for member in typ.__dict__.values():
        if hasattr(member, '__enforcer__'):
            raise AutoclassDecorationException('It seems that @runtime_validation decorator was applied to type <'
                                               + str(typ) + '> BEFORE ' + calling_decorator + '. This is not supported '
                                               'as it may lead to counter-intuitive behaviour, please change the order '
                                               'of the decorators on <' + str(typ) + '>')


def _call_decorator_with_or_without_args(decorator_function_impl: Callable, objectIsFunction: bool, *args, **kwargs):
    """
    Utility function for all decorators: handles the fact that 
    * if the decorator has been called with arguments, then we have to return a decorator factory
    * if the decorator has been called without arguments, then we have to directly return the decoration result

    :param decorator_function_impl:
    :param args:
    :param kwargs:
    :return:
    """

    if len(args) > 0 and ((objectIsFunction and callable(args[0])) or (not objectIsFunction and isclass(args[0]))):
        # we have been called without argument. In that case 'args[0]' contains the object_to_decorate
        return decorator_function_impl(*args, **kwargs)
    else:
        # called with arguments : lets return a factory function
        def f(object_to_decorate):
            return decorator_function_impl(object_to_decorate, *args, **kwargs)
        return f


def _create_function_decorator__robust_to_args(decorator_impl: Callable, *args, **kwargs):
    """
    Utility method to implement a decorator for functions based on the provided implementation and arguments
    
    :param decorator_impl: a function taking as first argument a function
    :param args: 
    :param kwargs: 
    :return: 
    """
    return _call_decorator_with_or_without_args(decorator_impl, True, *args, **kwargs)


def _create_class_decorator__robust_to_args(decorator_impl: Callable, *args, **kwargs):
    """
    Utility method to create a decorator for classes based on the provided implementation and arguments
    
    :param decorator_impl: 
    :param args: 
    :param kwargs: 
    :return: 
    """
    return _call_decorator_with_or_without_args(decorator_impl, False, *args, **kwargs)


def apply_on_func_args(func, cur_args, cur_kwargs,
                       signature_attrs, signature_defaults, signature_varargs, signature_varkw,
                       func_to_apply, func_to_apply_paramers_dict):
    """
    Applies func_to_apply on each argument of func according to what's received in current call (cur_args, cur_kwargs).
    For each argument of func named 'att' in its signature, the following method is called:

    `func_to_apply(cur_att_value, func_to_apply_paramers_dict[att], func, att_name)`

    :param func:
    :param cur_args:
    :param cur_kwargs:
    :param signature_attrs:
    :param signature_defaults:
    :param signature_varargs:
    :param signature_varkw:
    :param func_to_apply:
    :param func_to_apply_paramers_dict:
    :return:
    """
    # handle default values and kw arguments
    for attr, val in zip(reversed(signature_attrs), reversed(signature_defaults or [])):
        if attr in func_to_apply_paramers_dict.keys():
            # set default or provided value
            if attr in cur_kwargs.keys():
                # provided: we never seem to enter here, why ? maybe depends on the version of python
                func_to_apply(cur_kwargs[attr], func_to_apply_paramers_dict[attr], func, attr)
            else:
                # default
                func_to_apply(val, func_to_apply_paramers_dict[attr], func, attr)

    # handle positional arguments
    for attr, val in zip(signature_attrs, cur_args):
        if attr in func_to_apply_paramers_dict.keys():
            func_to_apply(val, func_to_apply_paramers_dict[attr], func, attr)

    # handle varargs : since we dont know their name, they can only be validated as a whole
    if signature_varargs:
        remaining_args = cur_args[len(signature_attrs):]
        if signature_varargs in func_to_apply_paramers_dict.keys():
            func_to_apply(remaining_args, func_to_apply_paramers_dict[signature_varargs], func, signature_varargs)

    # handle varkw : since we know their names, they can be validated either as a whole or independently
    if signature_varkw:
        # either the item is the whole dictionary (if func signature is generic such as in **kwargs)
        if signature_varkw in func_to_apply_paramers_dict.keys():
            # value = the cur_kwargs as a whole
            func_to_apply(cur_kwargs, func_to_apply_paramers_dict[signature_varkw], func, signature_varkw)
        else:
            # or each item is handled independently (if func signature contains the kw args names such as a, b)
            for attr, val in cur_kwargs.items():
                if attr in func_to_apply_paramers_dict.keys():
                    func_to_apply(val, func_to_apply_paramers_dict[attr], func, attr)
